
;Linux signal handling demonstration
;IA32 platform

;Constants =====================================================================
;Page protection constants
PROT_NONE=	0
PROT_READ=	1
PROT_WRITE=	2
PROT_EXEC=	4

;Sigaction flags
;defined in arch/x86/include/asm/signal.h
SA_NOCLDSTOP=	0x00000001	;turns off SIGCHLD when children stop
SA_NOCLDWAIT=	0x00000002	
SA_SIGINFO=		0x00000004
SA_ONSTACK=		0x08000000	;indicates that the registered stack_t will be used
SA_RESTART=		0x10000000	;flag to get restarting signals
SA_NODEFER=		0x40000000	;prevent the current signal from being masked 
							;in the handler
SA_RESETHAND=	0x80000000	;clears the handler when the signal is delivered
SA_NOMASK=		SA_NODEFER	;Historical Linux names for the Single Unix names
SA_ONESHOT=		SA_RESETHAND;NODEFER and RESETHAND respectively
SA_RESTORER=	0x04000000

;Sigset constants
_COMPAT_NSIG=		64
_COMPAT_NSIG_BPW=	32
_COMPAT_NSIG_WORDS= _COMPAT_NSIG / _COMPAT_NSIG_BPW

;Signal numbers
SIGSEGV=		0x0B

;Used system calls
__NR_exit=			0x01
__NR_write=			0x04
__NR_sigaction=		0x43
__NR_mprotect=		0x7D

;Structures ====================================================================
;struct sigcontext_ia32
;defined in arch/x86/include/asm/sigcontext32.h
struc	sigcontext_ia32
{
	;Field			size	offset	comment
	.gs				dw	?	;0x00
	.__gsh			dw	?	;0x02
	.fs				dw	?	;0x04
	.__fsh			dw	?	;0x06
	.es				dw	?	;0x08
	.__esg			dw	?	;0x0A
	.ds				dw	?	;0x0C
	.__dsh			dw	?	;0x0E
	.edi			dd	?	;0x10
	.esi			dd	?	;0x14
	.ebp			dd	?	;0x18
	.esp			dd	?	;0x1C
	.ebx			dd	?	;0x20
	.edx			dd	?	;0x24
	.ecx			dd	?	;0x28
	.eax			dd	?	;0x2C
	.trapno			dd	?	;0x30
	.err			dd	?	;0x34
	.eip			dd	?	;0x38
	.cs				dw	?	;0x3C
	.__csh			dw	?	;0x3E
	.flags			dd	?	;0x40
	.sp_at_signal	dd	?	;0x44
	.ss				dw	?	;0x48
	.__ssh			dw	?	;0x4A
	.fpstate		dd	?	;0x4C	Pointer to struct _fpstate_ia32
							;		which contains stored state of the FPU
							;		This structure may be found in
							;		arch/x86/include/asm/sigcontext32.h
	.oldmask		dd	?	;0x50
	.cr2			dd	?	;0x54
}

;typedef struct siginfo
;This structure is not fully covered here
;defined in include/asm-generic/siginfo.h
struc siginfo
{
	.si_signo		dd	?	;0x00
	.si_errno		dd	?	;0x04
	.si_code		dd	?	;0x08
	;unions depending on signal
}

;struct sigaltstack_ia32
;defined in arch/x86/include/asm/ia32.h
;Describes alternate stack if any
struc sigaltstack_ia32
{
	;Field 			size 	offset
	.ss_sp			dd	?	;0x00
	.ss_flags		dd	?	;0x04
	.ss_size		dd	?	;0x08
};stack_ia32_t

;typedef struct compat_sigset_t
;defined in include/linux/compat.h
;_COMPAT_NSIG_WORDS = (_COMPAT_NSIG / _COMPAT_NSIG_BPW)
;_COMPAT_NSIG and _COMPAT_NSIG_BPW defined in arch/x86/include/asm/compat.h
;Describes sigset mask
struc compat_sigset_t
{
	.sig			rd	_COMPAT_NSIG_WORDS	;=64/4
}

;struct ucontext_ia32
;defined in arch/x86/include/asm/ia32.h
;Pointer to this structure is passed to extended handler
struc ucontext_ia32
{
	;Field			size				offset
	.uc_flags		dd	?				;0x00
	.uc_link		dd	?				;0x04
	.uc_stack		sigaltstack_ia32	;0x08
	.uc_mcontext	sigcontext_ia32		;0x14
	.uc_sigmask		compat_sigset_t		;0x6C
}

;struct sigaction32
;defined in arch/x86/include/asm/ia32.h
;Describes sigaction structure
struc sigaction32
{
	;Field			size				offset
	.sa_handler		dd	?				;0x00	Address of the custom
										;		signal handler
	.sa_restorer	dd	?				;0x08	Address of the procedure which
										;		restores user context. It suggested
										;		not to use this on newer platforms.
										;		VDSO handles this perfectly.
	.sa_flags		dd	?				;0x04	Sigaction flags
	.sa_mask		dd	?		;0x0C	Signal mask
}


;============ ACTUAL CODE ======================================================

format ELF executable
entry	_start

;Code section ------------------------------------------------------------------
segment executable
	;Signal handler
	;void handler(int signum, siginfo_t* sinfo, void* context)
	_handler:
		push ebp
		mov ebp,esp
		push eax ebx ecx edx 
		call .get_eip
		
		;Get current EIP
	.get_eip:
		;mov ebx,[esp]
		pop ebx
		
		;Page align for mprotect
		and ebx,0xFFFFF000
		;Get the length of the region
		mov ecx,_start.hidden_code+_start.hidden_code_length
		sub ecx,ebx
		;Set protection
		mov edx,PROT_READ or PROT_WRITE or PROT_EXEC
		mov eax,__NR_mprotect
		;Execute call
		int 0x80
		
		;We have to check the result as we cannot proceed if the memory has not
		;become writable. If not, then we just exit
		or eax,0
		jnz unrecoverable_error
		
		;We need to check the reason for signal
		mov eax,[ebp+16] ;void* context
		
		virtual at eax
			.context	ucontext_ia32
		end virtual
		
		;compare the address of _start.reason1 with the contents of the EIP
		;in stored context
		cmp	dword[.context.uc_mcontext.eip],_start.reason1
		jnz ._reason2
		
		;this is the first segfault, so we need to decode stuff
		;save the address for mprotect and the address of ucontext_ia32
		push eax ebx
		;print string2
		mov eax,__NR_write
		mov ebx,1
		mov ecx,string2
		mov edx,string2_len
		int 0x80
		
		;restore the address for mprotect and the address of ucontext_ia32
		pop ebx eax
		;decode code
		call _handler.codec1
		;decode string1
		call _handler.codec2
		;set new EIP in the stored context
		mov dword[.context.uc_mcontext.eip],_start.hidden_code
		jmp .finish
		
	._reason2:	
		;this is the second segfault, so we are about to encode things
		;save the address for mprotect and the address of ucontext_ia32
		push eax ebx
		;print string3
		mov eax,__NR_write
		mov ebx,1
		mov ecx,string3
		mov edx,string3_len
		int 0x80
		;restore the address for mprotect and the address of ucontext_ia32
		pop ebx eax
		;encode string
		call _handler.codec2
		;encode code
		call _handler.codec1
		;set new EIP in the stored context
		mov dword[.context.uc_mcontext.eip],_start.finish	
		
	.finish:
		;restore protection attributes of the code section
		mov ecx,_start.hidden_code+_start.hidden_code_length
		sub ecx,ebx
		mov eax,__NR_mprotect
		mov edx,PROT_READ or PROT_EXEC
		int 0x80
		
		pop edx ecx ebx eax
		mov esp,ebp
		pop ebp
		ret
		
	.codec1:
		;work on code
		mov ecx,_start.hidden_code_length
		mov edx,_start.hidden_code
	.do1:
		xor byte[edx],0xE5
		inc edx
		loop .do1
		ret
		
	.codec2:
		;work on data
		push eax
		mov ecx,string1_len
		mov edx,string1
		;load key
		mov al,[_start.hidden_code]
		xor al,[_start.hidden_code+20]
	.do2:
		xor byte[edx],al
		inc edx
		loop .do2
		pop eax
		ret
		
	unrecoverable_error:
		mov eax,__NR_exit
		xor ebx,ebx
		int 0x80
		
	
		
	;Program's entry point
	;int main(int argc, char** argv, char** env)
	_start:
		;Setup sigaction
		; handler
		mov [action.sa_handler],_handler
		; flags
		; we use SA_SIGINFO as we are interested in extended information
		mov [action.sa_flags],SA_SIGINFO		
		;Call sigaction
		; sigaction system call number
		mov eax,__NR_sigaction
		; signal number
		mov ebx,SIGSEGV
		; address of sigaction structure 
		; kernel_sigaction, not the libc one
		mov ecx,action
		; address of sigaction structure
		; to store previous action
		xor edx,edx
		; call
		int 0x80
		
		;Cause segmentation fault to run the handler
		;and decrypt the following code
		mov eax,.hidden_code
	.reason1:
		xor dword [eax],0x8BADF00D
		
		;this portion of code will be encrypted in compilation time
		;using simple XOR algorithm
	.hidden_code:
		;print "It works!"
		mov eax,__NR_write
		mov ebx,1			;stdout
		mov ecx,string1		;string to print
		mov edx,string1_len ;length of the string
		int 0x80
		
		;Cause segmentation fault again to run the handler and 
		;encrypt back the code above
		mov eax,.hidden_code
	.reason2:
		xor dword [eax],0x8BADF00D
	.hidden_code_length = $ - .hidden_code
		
	.finish:
		mov eax,__NR_exit
		xor ebx,ebx
		int 0x80

		
;Encrypt the hidden code at compile time
repeat .hidden_code_length;
	load a byte from .hidden_code+%-1
	store byte a xor 0xE5 at .hidden_code+%-1
end repeat


;Data section ------------------------------------------------------------------
segment writeable
	action			sigaction32
	
	string1			db	'It works!',10
	string1_len = $ - string1

	string2			db	'We are going to decrypt some code here',10
	string2_len = $ - string2
	
	string3			db	'Hiding back the code',10
	string3_len = $ - string3

;Encrypt the string1 content at compile time
repeat string1_len
	load a byte from string1+%-1
	b = 0xB8 xor 0xCD
	store byte a xor b at string1+%-1
end repeat
