
;Linux signal handling demonstration
;IA64 platform


;Constants =====================================================================
;Page protection constants
PROT_NONE=	0
PROT_READ=	1
PROT_WRITE=	2
PROT_EXEC=	4

;Sigaction flags
;defined in arch/x86/include/asm/signal.h
SA_NOCLDSTOP=	0x00000001	;turns off SIGCHLD when children stop
SA_NOCLDWAIT=	0x00000002	
SA_SIGINFO=		0x00000004
SA_ONSTACK=		0x08000000	;indicates that the registered stack_t will be used
SA_RESTART=		0x10000000	;flag to get restarting signals
SA_NODEFER=		0x40000000	;prevent the current signal from being masked 
							;in the handler
SA_RESETHAND=	0x80000000	;clears the handler when the signal is delivered
SA_NOMASK=		SA_NODEFER	;Historical Linux names for the Single Unix names
SA_ONESHOT=		SA_RESETHAND;NODEFER and RESETHAND respectively
SA_RESTORER=	0x04000000

;Sigset constants
_NSIG=				64
_NSIG_BPW=			8
_NSIG_WORDS= _NSIG / _NSIG_BPW

;Signal numbers
SIGSEGV=			0x0B

;Used system calls
__NR_exit=			0x3C
__NR_write=			0x01
__NR_rt_sigaction=	0x0D
__NR_mprotect=		0x0A
__NR_rt_sigreturn=	0x0F

;Structures ====================================================================

;struct sigcontext
;defined in arch/x86/include/asm/sigcontext.h
struc sigcontext
{
	;Field			size	offset
	.r8				dq	?	;0x00
	.r9				dq	?	;0x08
	.r10			dq	?	;0x10
	.r11			dq	?	;0x18
	.r12			dq	?	;0x20
	.r13			dq	?	;0x28
	.r14			dq	?	;0x30
	.r15			dq	?	;0x38
	.rdi			dq	? 	;0x40
	.rsi			dq	?	;0x48
	.rbp			dq	?	;0x50
	.rbx			dq	?	;0x58
	.rdx			dq	?	;0x60
	.rax			dq	?	;0x68
	.rcx			dq	?	;0x70
	.rsp			dq	?	;0x78
	.rip			dq	?	;0x80
	.eflags			dq	?	;0x88
	.cs				dw	?	;0x90
	.gs				dw	?	;0x92
	.fs				dw	?	;0x94
	.__pad0			dw	?	;0x96
	.err			dq	?	;0x98
	.trapno			dq	?	;0xA0
	.oldmask		dq	?	;0xA8
	.cr2			dq	?	;0xB0
	.fpstate		dq	?	;0xB8
	.reserved		rq	8	;0xC0
}


struc sigaltstack
{
	.ss_sp			dq	?
	.ss_flags		dq	?
	.ss_size		dq	?
};stack_t

struc sigset_t
{
	.sig:
	repeat _NSIG_WORDS
		dq	?
	end repeat
}

;struct kernel_sigaction
;this struct was taken from the glibc sources
struc sigaction
{
	.__sigaction_handler	dq	?
	.sa_flags				dq	?
	.sa_restorer			dq	?
	.__val					sigset_t
}


struc ucontext
{
	.uc_flags				dq	?
	.uc_link				dq	?
	.uc_stack				sigaltstack
	.uc_mcontext			sigcontext
	.uc_sigmask				sigset_t
}


;============ ACTUAL CODE ======================================================

format ELF64 executable
entry	_start

;Code section ------------------------------------------------------------------
segment executable
	_restorer:
		mov rax,__NR_rt_sigreturn
		syscall

	_handler:
		push rdx
		call .get_rip
		;Get current RIP
	.get_rip:
		pop rdi
		;Page align for mprotect
		and di,0xF000
		mov rsi,_start.hidden_code+_start.hidden_code_length
		sub rsi,rdi
		mov rdx,PROT_READ or PROT_WRITE or PROT_EXEC
		mov rax,__NR_mprotect
		syscall
		;------
		pop rdx
		virtual at rdx
			.context	ucontext
		end virtual
		cmp qword[.context.uc_mcontext.rip],_start.reason1
		jnz .reason2
		push rdx rdi
		mov rax,__NR_write
		mov rdi,1
		mov rsi,string2
		mov rdx,string2_len
		syscall
		;------
		pop rdi rdx
		call _codec1
		call _codec2
		mov qword[.context.uc_mcontext.rip],_start.hidden_code
		jmp .finish
	.reason2:
		push rdx rdi
		mov rax,__NR_write
		mov rdi,1
		mov rsi,string3
		mov rdx,string3_len
		syscall
		;------
		pop rdi rdx
		call _codec2
		call _codec1
		mov qword[.context.uc_mcontext.rip],_start.exit
	.finish:
		mov rsi,_start.hidden_code+_start.hidden_code_length
		sub rsi,rdi
		mov rdx,PROT_READ or PROT_EXEC
		mov rax,__NR_mprotect
		syscall
		;------
		ret

	_codec1:	;code
		push rdi rcx
		mov rdi,_start.hidden_code
		mov rcx,_start.hidden_code_length
	.codec:
		xor byte[rdi],0xE5
		inc rdi
		loop .codec
		pop rcx rdi
		ret
	
	_codec2:	;data
		push rax rdi rcx
		mov al,[_start.hidden_code]
		xor al,[_start.hidden_code+15]
		mov rdi,string1
		mov rcx,string1_len
	.codec:
		xor [rdi],al
		inc rdi
		loop .codec
		pop rcx rdi rax
		ret

	_start:
		;Setup sigaction
		mov r10,_NSIG_WORDS
		mov [sa.__sigaction_handler],_handler
		mov [sa.sa_flags],SA_SIGINFO or SA_RESTORER
		mov [sa.sa_restorer],_restorer
		xor rdx,rdx
		mov rsi,sa
		mov	rdi,SIGSEGV
		mov rax,__NR_rt_sigaction
		syscall
		;------
		
		mov rax,.hidden_code
	.reason1:
		xor dword[rax],0xDEADBEEF

	.hidden_code:
		mov rax,__NR_write
		mov rdi,1
		mov rsi,string1
		mov rdx,string1_len
		syscall
		;------
		mov rax,.hidden_code
	.reason2:
		xor dword[rax],0xDEADBEEF
	.hidden_code_length = $ - .hidden_code


	.exit:
		mov rax,__NR_exit
		xor rdi,rdi
		syscall
		;------
		
;Encrypt the hidden code at compile time
repeat .hidden_code_length;
	load a byte from .hidden_code+%-1
	store byte a xor 0xE5 at .hidden_code+%-1
end repeat


;Data section ------------------------------------------------------------------
segment writeable
	sa				sigaction

	string1			db	'It works!',0x0A
	string1_len	= $ - string1
	
	string2			db	'We are about to decrypt some code',0x0A
	string2_len	= $ - string2

	string3			db	'Hiding back the code',0x0A
	string3_len = $ - string3

;Encrypt the string1 content at compile time
repeat string1_len
	load a byte from string1+%-1
	b = 0x48 xor 0xC7
	store byte a xor b at string1+%-1
end repeat
